/**
* @file
* Main application.
*
* This file contains the main application. The window is created and all user
* input is handled here. Starting of new processes and command recognition is
* also part of the main application.
*/

/*
 * Copyright (c) 2009, Felix Opatz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of  nor the names of its contributors may be used to
 *    endorse or promote products derived from this software without specific
 *    prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <assert.h>
#include <windows.h>
#include <tchar.h>

#include "AppLaunch.h"
#include "AppLaunchRes.h"
#include "Config.h"
#include "ErrorHandling.h"
#include "AutoComplete.h"
#include "StringUtil.h"

/****************************************************************************
* GLOBAL DEFINITIONS
*****************************************************************************/

/**
* Global variable to store the instance pointer.
*/
HINSTANCE g_hInstance;

/**
* Global variable holding the handle of the main window.
*/
HWND g_hMainWindow;

/****************************************************************************
* FILE LOCAL DEFINITIONS
*****************************************************************************/

/**
* Width of the main window
*/
#define WND_WIDTH            320

/**
* Height of the main window
*/
#define WND_HEIGHT           120

/**
* Size for the big font used to indicate user input
*/
#define BIG_FONT_SIZE        36

/**
* Size for the small font used to draw the splash screen
*/
#define SMALL_FONT_SIZE      18

/**
* Maximum length of the HotKey string used in splash screen.
*/
#define MAX_HOT_KEY_STR_LEN  100

/**
* Maximum length of commands that can be entered by the user.
*/
#define MAX_CMD_LEN          8192

/**
* Name of the kernel object used as instance marker.
*/
#define SINGLE_INST_EVENT    (APP_NAME TEXT(".SingleInstance"))

/**
* IDs for the HotKey registration
*/
#define ID_HOTKEY            100
#define ID_HOTKEY_ALT        101

/**
* Flag to remember if alternative ID is used
*/
static BOOL fHotKeyAltIdUsed;

/**
* ID for Splash Screen timer
*/
#define ID_SPLASH_TIMER      102

/**
* Window message to reshow splash screen
*/
#define WMU_SHOW_SPLASH      (WM_USER + 1)

/**
* Format used for clipboard text.
*/
#ifdef UNICODE
#define CF_TCHAR CF_UNICODETEXT
#else
#define CF_TCHAR CF_TEXT
#endif

/**
* Version as string
*/
#define APP_VERSION_STRING TEXT("1.3")

/**
* Copyright information for splash screen
*/
#define COPYRITH_INFO TEXT("(C) 2009 by Felix Opatz")

/**
* URL to project page
*/
#define PROJECT_URL TEXT("http://www.zotteljedi.de/")

/**
* Flag that stores the current visibility state of the mian window.
*/
static BOOL g_fMainWindowVisible;

/**
* Pointer to the application's configuration.
*/
static CONFIG_DATA *g_lpAppConfig;

/**
* Pointer to path of current configuration file
*/
static LPCTSTR g_lpConfigPath;

/**
* Current command entered by user.
*/
static TCHAR g_szCurrCmd[MAX_CMD_LEN];

/**
* Length of current command.
*/
static DWORD g_dwCurrCmdLen;

/**
* Flags that indicate the result of a command execution.
*/
#define EX_FLAG_FAILURE   0x01  /** Execution failed        */
#define EX_FLAG_HIDE      0x02  /** Hide main window        */
#define EX_FLAG_BEEP      0x04  /** Beep (invalid command)  */
#define EX_FLAG_CLEAR     0x08  /** Clear text input        */

/****************************************************************************
* FILE LOCAL FUNCTIONS
*****************************************************************************/

/**
* Draws the splash screen into the given device context.
* @param hWnd Handle to window to paint into
* @param hdc Device context to draw to
*/
static void DrawSplashScreen(HWND hWnd, HDC hdc)
{
	RECT rect;
	LPCTSTR lpAppNameAndVersion = APP_NAME TEXT(", Version ") APP_VERSION_STRING;
	LPCTSTR lpCopyrightInfo = COPYRITH_INFO;
	LPCTSTR lpProjectUrl = PROJECT_URL;
	LPCTSTR lpHotKey = TEXT("HotKey: ");
	TCHAR *lpSplashScreenContent = NULL;
	int iResultLen;
	DWORD dwSplashContentCch;
	TCHAR szHotKeyAsString[MAX_HOT_KEY_STR_LEN];
	
	if (GetHotKeyAsString(szHotKeyAsString, sizeof(szHotKeyAsString),
		g_lpAppConfig->uHotKeyModifiers, g_lpAppConfig->uHotKeyVirtKey) == 0)
	{
		// Shall never happen since config must be valid
		_tcscpy(szHotKeyAsString, TEXT("<not configured>"));
	}

	dwSplashContentCch = TCSLEN(lpAppNameAndVersion) + TCSLEN(lpCopyrightInfo) +
		TCSLEN(lpProjectUrl) + TCSLEN(lpHotKey) + TCSLEN(szHotKeyAsString) + 20;
	lpSplashScreenContent = malloc(dwSplashContentCch * sizeof(TCHAR));

	// Out of memory
	if (lpSplashScreenContent == NULL)
		return;

	iResultLen = _sntprintf(lpSplashScreenContent, dwSplashContentCch, TEXT("%s\n%s\n%s%\n%s%s"),
		lpAppNameAndVersion, lpCopyrightInfo, lpProjectUrl, lpHotKey, szHotKeyAsString);
	lpSplashScreenContent[iResultLen] = '\0';

	GetClientRect(hWnd, &rect);

	rect.top += ((rect.bottom - rect.top) - (4 * SMALL_FONT_SIZE)) / 2;

	SetTextColor(hdc, g_lpAppConfig->crTextColor);
	DrawText(hdc, lpSplashScreenContent, iResultLen, &rect, DT_CENTER);
}

/**
* Draws the currently entered text into the given device context.
* @param hWnd Handle to window to paint into
* @param hdc Device context to draw to
*/
static void DrawCurrentText(HWND hWnd, HDC hdc)
{
	SIZE size;
	RECT rect;
	UINT uFormat;

	GetClientRect(hWnd, &rect);
	GetTextExtentPoint32(hdc, g_szCurrCmd, g_dwCurrCmdLen, &size);

	if (size.cx >= WND_WIDTH)
		uFormat = DT_RIGHT;
	else
		uFormat = DT_CENTER;

	uFormat |= DT_VCENTER | DT_SINGLELINE;

	SetTextColor(hdc, g_lpAppConfig->crTextColor);
	DrawText(hdc, g_szCurrCmd, g_dwCurrCmdLen, &rect, uFormat);
}


/**
* Shows or hides the main window
* @param fShow Flag of visibility-state
*/
static void ShowMainWindow(BOOL fShow)
{
	if (fShow == TRUE)
	{
		ShowWindow(g_hMainWindow, SW_SHOW);
		SetForegroundWindow(g_hMainWindow);
		g_fMainWindowVisible = TRUE;
	}
	else
	{
		ShowWindow(g_hMainWindow, SW_HIDE);
		g_fMainWindowVisible = FALSE;
	}
}

/**
* Reload current configuration
* @return TRUE if reload was successful. If FALSE is returned, loading the new
* configuration failed and the old configuration is still in use.
*/
static BOOL ReloadConfig(void)
{
	CONFIG_DATA *lpNewConfig;
	BOOL fHotKeyRegistered = FALSE;
	BOOL fUnregisterHotKey = FALSE;
	int iOldHotKeyId;
	int iNewHotKeyId;
	LOGBRUSH lb;
	HBRUSH hOldBrush;
	HBRUSH hNewBrush;

	lpNewConfig = calloc(1, sizeof(CONFIG_DATA));
	if (lpNewConfig == NULL)
		return FALSE;

	// Try to load configuration from file
	if (LoadConfig(g_lpConfigPath, lpNewConfig) == FALSE)
	{
		free(lpNewConfig);
		return FALSE;
	}

	iOldHotKeyId = fHotKeyAltIdUsed ? ID_HOTKEY_ALT : ID_HOTKEY;
	iNewHotKeyId = fHotKeyAltIdUsed ? ID_HOTKEY : ID_HOTKEY_ALT;

	// Try to register new HotKey before unregistering old HotKey
	if ((lpNewConfig->uHotKeyModifiers != INVALID_HOT_KEY_MODIFIER) &&
		(lpNewConfig->uHotKeyVirtKey != INVALID_HOT_KEY_VIRT_KEY))
	{
		if ((lpNewConfig->uHotKeyModifiers != g_lpAppConfig->uHotKeyModifiers) ||
			(lpNewConfig->uHotKeyVirtKey != g_lpAppConfig->uHotKeyVirtKey))
		{
			fHotKeyRegistered = RegisterHotKey(g_hMainWindow, iNewHotKeyId,
				lpNewConfig->uHotKeyModifiers, lpNewConfig->uHotKeyVirtKey);
			if (fHotKeyRegistered)
			{
				// Toggle HotKey-ID alternation
				fHotKeyAltIdUsed = !fHotKeyAltIdUsed;
				// New HotKey registered, unregister old HotKey
				fUnregisterHotKey = TRUE;
			}
		}
		else
		{
			// Registration not needed, thus successful
			fHotKeyRegistered = TRUE;
			// New HotKey is same as old HotKey, do not unregister
			fUnregisterHotKey = FALSE;
		}
	}

	if (fHotKeyRegistered == FALSE)
	{
		free(lpNewConfig);
		ReportError(ERROR_REGISTER_HOTKEY_FAILED);
		return FALSE;
	}

	// Unregister hot key
	if (fUnregisterHotKey)
		UnregisterHotKey(g_hMainWindow, iOldHotKeyId);

	// Check if window background color has changed
	if (g_lpAppConfig->crBackgroundColor != lpNewConfig->crBackgroundColor)
	{
		// Create new brush for window background
		lb.lbColor = lpNewConfig->crBackgroundColor;
		lb.lbStyle = BS_SOLID;
		hNewBrush = CreateBrushIndirect(&lb);
		
		// Switch brushes
		hOldBrush = (HBRUSH)GetClassLong(g_hMainWindow, GCL_HBRBACKGROUND);
		SetClassLong(g_hMainWindow, GCL_HBRBACKGROUND, (LONG)hNewBrush);

		DeleteObject(hOldBrush);
	}

	// Free memory used to hold configuration
	FreeConfig(g_lpAppConfig);
	free(g_lpAppConfig);

	// Switch pointer to new configuration
	g_lpAppConfig = lpNewConfig;

	return TRUE;
}

/**
* Constructs a command line from the path and options/arguments and then starts
* a process in the specified working directory. Options and arguments are
* appended to the command line, in this order.
* @param lpPath Path to executable to start
* @param lpDir Working directory (optional, may be NULL)
* @param lpOptions Options for command line (may be NULL)
* @param lpArguments More options for command line (may be NULL)
* @return TRUE if starting process was successful
*/
static BOOL StartProcess(LPCTSTR lpPath, LPCTSTR lpDir, LPCTSTR lpOptions, LPCTSTR lpArguments)
{
	BOOL fSucceeded;
	TCHAR *lpCommandLine;
	DWORD dwCommandLineLenCch;
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	int iStrPos = 0;

	dwCommandLineLenCch = 0;

	dwCommandLineLenCch += (lpPath != NULL) ? TCSLEN(lpPath) : 0;
	dwCommandLineLenCch += (lpOptions != NULL) ? TCSLEN(lpOptions) : 0;
	dwCommandLineLenCch += (lpArguments != NULL) ? TCSLEN(lpArguments) : 0;
	dwCommandLineLenCch += 10;
	lpCommandLine = malloc(dwCommandLineLenCch * sizeof(TCHAR));

	// Out of memory
	if (lpCommandLine == NULL)
		return FALSE;

	*lpCommandLine = '\0';

	iStrPos += _sntprintf(lpCommandLine + iStrPos, dwCommandLineLenCch - iStrPos, TEXT("%s"), lpPath);
	if (lpOptions != NULL)
		iStrPos += _sntprintf(lpCommandLine + iStrPos, dwCommandLineLenCch - iStrPos, TEXT(" %s"), lpOptions);
	if (lpArguments != NULL)
		iStrPos += _sntprintf(lpCommandLine + iStrPos, dwCommandLineLenCch - iStrPos, TEXT(" %s"), lpArguments);

	ZeroMemory(&si, sizeof(si));
	si.cb = sizeof(si);
	ZeroMemory(&pi, sizeof(pi));
	fSucceeded = CreateProcess(NULL, lpCommandLine, NULL, NULL, FALSE, 0, NULL, lpDir, &si, &pi);

	if (fSucceeded == FALSE)
		ReportError(ERROR_CREATING_PROCESS, lpPath, GetLastError());

	free(lpCommandLine);

	return fSucceeded;
}

/**
* Executes the given command by starting a new process.
* The command is parsed into the name and parameters and then looked up
* against the command list.
* @param lpCommand Command to execute
* @return Error code indication success of execution, combination
* of EX_ flags.
*/
static unsigned int ExecuteCommand(LPCTSTR lpCommand)
{
	TCHAR *lpCommandName = NULL;
	const TCHAR *lpStart = lpCommand;
	DWORD dwCommandNameLenCch;
	COMMAND_ENTRY *lpCommandEntry = NULL;
	DWORD dwCommandIndex;
	BOOL fSucceeded = FALSE;
	unsigned int uResult = 0;

	// Skip whitespace
	while ((*lpCommand != '\0') && (_istspace(*lpCommand) != 0))
		lpCommand++;

	// Start of command name is here
	lpStart = lpCommand;

	// find end of command
	while ((*lpCommand != '\0') && (_istspace(*lpCommand) == 0))
		lpCommand++;

	dwCommandNameLenCch = (DWORD)(lpCommand - lpStart);

	// No command entered
	if (dwCommandNameLenCch == 0)
		return EX_FLAG_FAILURE | EX_FLAG_BEEP;

	lpCommandName = malloc((dwCommandNameLenCch + 1) * sizeof(TCHAR));

	// No memory left
	if (lpCommandName == NULL)
		return EX_FLAG_FAILURE | EX_FLAG_BEEP;

	memcpy(lpCommandName, lpStart, dwCommandNameLenCch * sizeof(TCHAR));
	lpCommandName[dwCommandNameLenCch] = '\0';

	while ((*lpCommand != '\0') && (_istspace(*lpCommand)))
		lpCommand++;

	// lpCommand now points to the options (if any)

	// Look up command
	for (dwCommandIndex = 0; dwCommandIndex < g_lpAppConfig->dwCommandCount; dwCommandIndex++)
		if (EQUALS(g_lpAppConfig->Commands[dwCommandIndex].lpName, lpCommandName))
			lpCommandEntry = g_lpAppConfig->Commands + dwCommandIndex;

	// Command found?
	if (lpCommandEntry != NULL)
	{
		uResult = EX_FLAG_CLEAR;

		switch (lpCommandEntry->iCmdType)
		{
		case CMD_TYPE_INTERNAL_QUIT:
			PostQuitMessage(0);
			uResult |= EX_FLAG_HIDE;
			break;
		case CMD_TYPE_INTERNAL_RELOAD:
			if (ReloadConfig())
				SendMessage(g_hMainWindow, WMU_SHOW_SPLASH, 0, 0);
			else
				uResult |= EX_FLAG_FAILURE;
			break;
		case CMD_TYPE_EXTERNAL:
			fSucceeded = StartProcess(lpCommandEntry->lpPath, lpCommandEntry->lpWorkingDir,
				lpCommandEntry->lpOptions, (TCSLEN(lpCommand) > 0) ? lpCommand : NULL);
			if (fSucceeded)
				uResult |= EX_FLAG_HIDE;
			else
				uResult |= EX_FLAG_FAILURE | EX_FLAG_BEEP;
			break;
		default:
			assert(0); // Unhandled command type
			uResult = EX_FLAG_FAILURE | EX_FLAG_BEEP;
			break;
		}
	}
	else
	{
		// Command not recognized
		uResult = EX_FLAG_FAILURE | EX_FLAG_BEEP;
	}

	free(lpCommandName);

	return uResult;
}

/**
* Generates a beep from the system speaker to indicate invalid input
*/
static void InvalidInputBeep(void)
{
	Beep(440, 100);
}

/**
* Inserts the text that is stored in the clipboard. If the text has multiple
* lines the insertion stops at the end of the first line.
* @return TRUE if insertion was successful (e.g. data was available)
*/
static BOOL InsertTextFromClipboard(void)
{
	HGLOBAL hClipboardData;
	LPCTSTR lpInsertString;
	DWORD dwMaxCharsToCopy;

	if (IsClipboardFormatAvailable(CF_TCHAR) == FALSE)
		return FALSE;

	if (OpenClipboard(g_hMainWindow) == FALSE)
		return FALSE;

	hClipboardData = GetClipboardData(CF_TCHAR);

	if (hClipboardData != NULL)
	{
		lpInsertString = GlobalLock(hClipboardData);
		if (lpInsertString != NULL)
		{
			dwMaxCharsToCopy = MAX_CMD_LEN - g_dwCurrCmdLen - 1;

			while ((*lpInsertString != '\0') && (dwMaxCharsToCopy > 0) &&
				(*lpInsertString != TEXT('\r')) && (*lpInsertString != TEXT('\n')))
			{
				g_szCurrCmd[g_dwCurrCmdLen] = *lpInsertString;
				g_dwCurrCmdLen++;
				lpInsertString++;
				dwMaxCharsToCopy--;
			}
			g_szCurrCmd[g_dwCurrCmdLen] = '\0';
			GlobalUnlock(hClipboardData);
		}
	}
	CloseClipboard();

	return TRUE;
}

/**
* WindowProc of main window.
* @param hWnd Window the message was received for
* @param uMsg Message received
* @param wParam WParam of message
* @param lParam LParam of message
* @return Result of message handling
*/
static LRESULT CALLBACK MainWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	static HFONT hBigFont;
	static HFONT hSmallFont;
	HFONT hOldFont;
	MINMAXINFO *lpMinMax;
	static BOOL fShowSplashScreen;
	HDC hdc;
	PAINTSTRUCT ps;
	BOOL fKeyConsumed;
	TCHAR szCompletedCmd[MAX_CMD_LEN];
	enum AutoCompletionResults CompletionResult;
	int iDropCount;
	int iDropIndex;
	DWORD dwMaxCharsLeft;
	static UINT_PTR uSplashTimer;
	unsigned int uExResult;

	switch (uMsg)
	{
	case WM_CREATE:
		hBigFont = CreateFont(BIG_FONT_SIZE, 0, 0, 0, FW_BOLD, FALSE, FALSE, FALSE,
			DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
			DEFAULT_QUALITY, FIXED_PITCH | FF_MODERN, NULL);

		hSmallFont = CreateFont(SMALL_FONT_SIZE, 0, 0, 0, FW_BOLD, FALSE, FALSE, FALSE,
			DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
			DEFAULT_QUALITY, FIXED_PITCH | FF_MODERN, NULL);

		fShowSplashScreen = TRUE;
		DragAcceptFiles(hWnd, TRUE);
		g_hMainWindow = hWnd;
		ShowMainWindow(TRUE);
		return 0;

	case WMU_SHOW_SPLASH:
		if (uSplashTimer != 0)
		{
			KillTimer(hWnd, uSplashTimer);
			uSplashTimer = 0;
		}
		fShowSplashScreen = TRUE;
		InvalidateRect(hWnd, NULL, TRUE);
		break;

	case WM_GETMINMAXINFO:
		lpMinMax = (MINMAXINFO*) lParam;
		lpMinMax->ptMaxSize.x = WND_WIDTH;
		lpMinMax->ptMaxSize.y = WND_HEIGHT;
		lpMinMax->ptMinTrackSize.x = WND_WIDTH;
		lpMinMax->ptMinTrackSize.y = WND_HEIGHT;
		lpMinMax->ptMaxTrackSize.x = WND_WIDTH;
		lpMinMax->ptMaxTrackSize.y = WND_HEIGHT;
		return 0;

	case WM_PAINT:
		hdc = BeginPaint(hWnd, &ps);
		SetBkColor(hdc, g_lpAppConfig->crBackgroundColor);

		if (fShowSplashScreen == TRUE)
		{
			hOldFont = SelectObject(hdc, hSmallFont);
			DrawSplashScreen(hWnd, hdc);

			// If visible time is 0 the splash screen will
			// be visible until the user dismisses it
			if ((uSplashTimer == 0) && (g_lpAppConfig->dwSplashVisibleTime > 0))
				uSplashTimer = SetTimer(hWnd, ID_SPLASH_TIMER,
					g_lpAppConfig->dwSplashVisibleTime * 1000, NULL);
		}
		else
		{
			hOldFont = SelectObject(hdc, hBigFont);
			DrawCurrentText(hWnd, hdc);
		}

		SelectObject(hdc, hOldFont);
		EndPaint(hWnd, &ps);
		return 0;

	case WM_HOTKEY:
		fShowSplashScreen = FALSE;
		// Invert current visibility
		ShowMainWindow(!g_fMainWindowVisible);
		return 0;

	case WM_CHAR:
		fShowSplashScreen = FALSE;
		
		fKeyConsumed = FALSE;
		switch (wParam)
		{
		case VK_ESCAPE:
			ShowMainWindow(FALSE);
			g_dwCurrCmdLen = 0;
			g_szCurrCmd[0] = '\0';
			fKeyConsumed = TRUE;
			return 0;

		case VK_BACK:
			if (g_dwCurrCmdLen > 0)
			{
				g_dwCurrCmdLen--;
				g_szCurrCmd[g_dwCurrCmdLen] = '\0';
				fKeyConsumed = TRUE;
			}
			break;

		case VK_TAB:
			fKeyConsumed = TRUE;

			CompletionResult = CompleteCommand(g_szCurrCmd, szCompletedCmd,
				sizeof(szCompletedCmd), g_lpAppConfig);

			switch (CompletionResult)
			{
			case CR_FAILED:
				InvalidInputBeep();
				break;

			case CR_PARTLY:
				InvalidInputBeep();
				// Fallthrough
			case CR_EXACT:
				_tcscpy(g_szCurrCmd, szCompletedCmd);
				g_dwCurrCmdLen = TCSLEN(g_szCurrCmd);
				break;
			}
			
			break;

		case VK_RETURN:
			fKeyConsumed = TRUE;
			uExResult = ExecuteCommand(g_szCurrCmd);

			if (uExResult & EX_FLAG_BEEP)
				InvalidInputBeep();
			if (uExResult & EX_FLAG_HIDE)
				ShowMainWindow(FALSE);
			if (uExResult & EX_FLAG_CLEAR)
			{
				g_dwCurrCmdLen = 0;
				g_szCurrCmd[0] = '\0';
			}
			break;

		default:
			if (g_dwCurrCmdLen >= MAX_CMD_LEN - 1)
				break;

			g_szCurrCmd[g_dwCurrCmdLen] = (TCHAR) wParam;
			g_dwCurrCmdLen++;
			g_szCurrCmd[g_dwCurrCmdLen] = '\0';
			fKeyConsumed = TRUE;

			break;
		}

		if (fKeyConsumed)
			InvalidateRect(hWnd, NULL, TRUE);

		break;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDM_PASTE)
		{
			if (InsertTextFromClipboard() == TRUE)
			{
				fShowSplashScreen = FALSE;
				InvalidateRect(hWnd, NULL, TRUE);
			}
			return 0;
		}
		break;

	case WM_RBUTTONDOWN:
		if (InsertTextFromClipboard() == TRUE)
		{
			fShowSplashScreen = FALSE;
			InvalidateRect(hWnd, NULL, TRUE);
			return 0;
		}
		break;

	case WM_DROPFILES:
		iDropCount = DragQueryFile((HDROP)wParam, 0xFFFFFFFF, NULL, 0);
		dwMaxCharsLeft = MAX_CMD_LEN - g_dwCurrCmdLen;
		for (iDropIndex = 0; (iDropIndex < iDropCount) && (dwMaxCharsLeft > 1); iDropIndex++)
		{
			DragQueryFile((HDROP)wParam, iDropIndex, g_szCurrCmd + g_dwCurrCmdLen, dwMaxCharsLeft);
			dwMaxCharsLeft -= TCSLEN(g_szCurrCmd + g_dwCurrCmdLen);
			g_dwCurrCmdLen += TCSLEN(g_szCurrCmd + g_dwCurrCmdLen);
		}
		g_szCurrCmd[g_dwCurrCmdLen] = '\0';
		fShowSplashScreen = FALSE;
		InvalidateRect(hWnd, NULL, TRUE);
		break;

	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;

	case WM_TIMER:
		if ((uSplashTimer != 0) && (wParam == uSplashTimer))
		{
			KillTimer(hWnd, uSplashTimer);
			uSplashTimer = 0;
			if (   (fShowSplashScreen == TRUE)
			    && IsHotKeyValid(g_lpAppConfig->uHotKeyModifiers, g_lpAppConfig->uHotKeyVirtKey))
			{
				fShowSplashScreen = FALSE;
				ShowMainWindow(FALSE);
			}
		}
		return 0;
	}

	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

/**
* Creates the main window by defining a custom window class. The new window
* is placed centered on the primary display and invisible.
* Note that this function leaks a brush handle (for the window background
* color), which is not a real problem because the main window is constructed
* once and persists until the application is closed.
* @param crBackgroundColor Color for the brush used to paint window background
* @return Handle to the created window, NULL if creation fails
*/
static HWND CreateMainWindow(const COLORREF crBackgroundColor)
{
	WNDCLASS wc;
	LOGBRUSH lb;
	HBRUSH hBrush;
	int iScreenWidth;
	int iScreenHeight;
	HWND hWnd;

	// Create brush for window background
	lb.lbColor = crBackgroundColor;
	lb.lbStyle = BS_SOLID;
	hBrush = CreateBrushIndirect(&lb);

	// Create custom window lcass
	ZeroMemory(&wc, sizeof(wc));
	wc.hInstance = g_hInstance;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.lpfnWndProc = MainWindowProc;
	wc.lpszClassName = TEXT("MainWindow");
	wc.hbrBackground = hBrush;
	wc.style = CS_HREDRAW | CS_VREDRAW;
	if (RegisterClass(&wc) == 0)
		return NULL;

	iScreenWidth = GetSystemMetrics(SM_CXSCREEN);
	iScreenHeight = GetSystemMetrics(SM_CYSCREEN);

	hWnd = CreateWindow(TEXT("MainWindow"), APP_NAME, WS_POPUP | WS_THICKFRAME,
		(iScreenWidth - WND_WIDTH) / 2, (iScreenHeight - WND_HEIGHT) / 2,
		WND_WIDTH, WND_HEIGHT,
		NULL, NULL, g_hInstance, NULL);

	return hWnd;
}

/**
* Extracts the command line parameters passed to the application to specify
* an alternative path of the configuration file.
* @param lpPathBuffer Pointer to a buffer that receives the path
* @param dwBufferLen Length of buffer in characters
*/
static void GetConfigPathFromCmdLine(LPTSTR lpBuffer, DWORD nSize)
{
	LPTSTR lpCmdLine;

	lpCmdLine = GetCommandLine();

	// Skip program name
	if (*lpCmdLine == TEXT('"'))
	{
		// Program name is enclosed by quoting chars
		lpCmdLine++;
		while ((*lpCmdLine != '\0') && (*lpCmdLine != TEXT('"')))
			lpCmdLine++;
		if (*lpCmdLine == TEXT('"'))
			lpCmdLine++;
	}
	else
	{
		// Program name is terminated with next whitespace char
		while ((*lpCmdLine != '\0') && (_istspace(*lpCmdLine) == 0))
			lpCmdLine++;
	}


	// Skip whitespace
	while ((*lpCmdLine != '\0') && (_istspace(*lpCmdLine) != 0))
		lpCmdLine++;

	// Copy next argument
	if (*lpCmdLine == TEXT('"'))
	{
		// Argument name is enclosed by quoting chars
		lpCmdLine++;
		while ((*lpCmdLine != '\0') && (*lpCmdLine != TEXT('"')) && nSize > 1)
		{
			*lpBuffer = *lpCmdLine;
			lpBuffer++;
			lpCmdLine++;
			nSize--;
		}
		if (*lpCmdLine == TEXT('"'))
			lpCmdLine++;
	}
	else
	{
		// Argument is terminated with next whitespace char
		while ((*lpCmdLine != '\0') && (_istspace(*lpCmdLine) == 0) && nSize > 1)
		{
			*lpBuffer = *lpCmdLine;
			lpBuffer++;
			lpCmdLine++;
			nSize--;
		}
	}

	// Terminate buffer
	*lpBuffer = '\0';
}

/****************************************************************************
* GLOBAL FUNCTIONS
*****************************************************************************/

// Disable warning C4100 "unreferenced formal parameter"
#pragma warning (disable: 4100)

/**
* Main function where execution starts. The command line is investigated and
* the specified configuration file is loaded. After that the main window is
* created, the HotKey is registered and the key accelerators are loaded. Then
* the message loop takes over control. When the message loop finished some
* cleanup actions take place and the application is left.
* @param hInstance Instance pointer
* @param hPrevInstance NULL (for compatibility)
* @param lpCmdLine Pointer to command line (ANSI only)
* @param nCmdShow Recommendation of how the main window should be shown
* @return Return value passed to operating system.
*   EXIT_FAILURE if an error happened, EXIT_SUCCESS if everything is fine.
*/
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	TCHAR szConfigPath[MAX_PATH + 1];
	MSG msg;
	HANDLE hEvent;
	HACCEL hAccel;
	BOOL fHotKeyRegistered = FALSE;

	// Check if this is the only instance running
	hEvent = CreateEvent(NULL, FALSE, FALSE, SINGLE_INST_EVENT);

	if (hEvent == NULL || GetLastError() == ERROR_ALREADY_EXISTS)
	{
		ReportError(ERROR_APP_ALREADY_STARTED);
		return EXIT_FAILURE;
	}

	// Store global pointers to shared resources
	g_hInstance = hInstance;
	g_lpConfigPath = szConfigPath;
	g_lpAppConfig = calloc(1, sizeof(CONFIG_DATA));

	if (g_lpAppConfig == NULL)
	{
		ReportError(ERROR_NO_MEMORY);
		return EXIT_FAILURE;
	}

	// lpCmdLine is an ANSI string. It is only investigated to find out if
	// there IS an argument specifying an alternative configuration file.
	// To actually get this path the GetCommandLine function is used.
	if (strlen(lpCmdLine) > 0)
	{
		// Get config path
		GetConfigPathFromCmdLine(szConfigPath, sizeof(szConfigPath) / sizeof(TCHAR));
	}
	else
	{
		// Get path of configuration file
		if (GetConfigPath(szConfigPath, sizeof(szConfigPath) / sizeof(TCHAR)) == FALSE)
			return EXIT_FAILURE;
	}

	// Try to load configuration from file
	if (LoadConfig(g_lpConfigPath, g_lpAppConfig) == FALSE)
		return EXIT_FAILURE;

	// Create the main window
	CreateMainWindow(g_lpAppConfig->crBackgroundColor);

	// Check if hot key is configured and register it
	if ((g_lpAppConfig->uHotKeyModifiers != INVALID_HOT_KEY_MODIFIER) &&
		(g_lpAppConfig->uHotKeyVirtKey != INVALID_HOT_KEY_VIRT_KEY))
	{
		fHotKeyRegistered = RegisterHotKey(g_hMainWindow, ID_HOTKEY,
			g_lpAppConfig->uHotKeyModifiers, g_lpAppConfig->uHotKeyVirtKey);
	}

	if (fHotKeyRegistered == FALSE)
	{
		ReportError(ERROR_REGISTER_HOTKEY_FAILED);
		return EXIT_FAILURE;
	}

	// Original HotKey-ID is used on startup
	fHotKeyAltIdUsed = FALSE;

	// Load accelerators for the paste function
	hAccel = LoadAccelerators(g_hInstance, TEXT("KeyAccel"));

	// Main message loop
	while (GetMessage(&msg, NULL, 0, 0))
	{
		if (TranslateAccelerator(g_hMainWindow, hAccel, &msg) == 0)
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	// Unregister hot key
	UnregisterHotKey(g_hMainWindow, fHotKeyAltIdUsed ? ID_HOTKEY_ALT : ID_HOTKEY);

	// Free memory used to hold configuration
	FreeConfig(g_lpAppConfig);
	free(g_lpAppConfig);
	g_lpAppConfig = NULL;

	return EXIT_SUCCESS;
}
