/**
* @file
* Auto completion functionality. A fragment of a command can be passed and 
* next unique part is determined.
*/

/*
 * Copyright (c) 2009, Felix Opatz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of  nor the names of its contributors may be used to
 *    endorse or promote products derived from this software without specific
 *    prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <windows.h>
#include <tchar.h>

#include "Config.h"
#include "AppLaunch.h"
#include "AutoComplete.h"
#include "StringUtil.h"

/**
* Shortens String A to be the largest possible prefix of String B
* (ignoring case)
* @param lpStringA Pointer to String A (writable, will be truncated)
* @param lpStringB Pointer to String B (readonly)
*/
static void ReduceStringToUniquePart(LPTSTR lpStringA, LPCTSTR lpStringB)
{
	while ((*lpStringA != 0) && (_totlower(*lpStringA) == _totlower(*lpStringB)))
	{
		lpStringA++;
		lpStringB++;
	}

	*lpStringA = '\0';
}

/**
* Completes a command name fragment to either match an existing command or
* to match the next common part.
* @param lpCommand Command name to complete
* @param lpBuffer Pointer to buffer to receive completed command name
* @param dwBufferLen Size of buffer in characters
* @param lpConfig Pointer to configuration data
* @return Result of completion attempt
*/
enum AutoCompletionResults
CompleteCommand(LPCTSTR lpCommand, LPTSTR lpBuffer, DWORD dwBufferLen, const CONFIG_DATA *lpConfig)
{
	DWORD dwEntryIndex = 0;
	TCHAR *lpBestMatch = NULL;
	BOOL fExactMatch = FALSE;

	// Find first match
	for (dwEntryIndex = 0; dwEntryIndex < lpConfig->dwCommandCount; dwEntryIndex++)
		if (PREFIX(lpCommand, lpConfig->Commands[dwEntryIndex].lpName))
		{
			lpBestMatch = _tcsdup(lpConfig->Commands[dwEntryIndex].lpName);
			fExactMatch = TRUE;
			break;
		}

	// Prefix doesn't match any commands
	if (lpBestMatch == NULL)
		return CR_FAILED;

	// Find the largest unique part
	while ((dwEntryIndex < lpConfig->dwCommandCount - 1) &&
		(PREFIX(lpCommand, lpConfig->Commands[dwEntryIndex + 1].lpName)))
	{
		fExactMatch = FALSE;
		if (PREFIX(lpBestMatch, lpConfig->Commands[dwEntryIndex + 1].lpName) == FALSE)
			ReduceStringToUniquePart(lpBestMatch, lpConfig->Commands[dwEntryIndex + 1].lpName);
		dwEntryIndex++;
	}

	// If length is equal there was nothing completed, so the operation
	// is considered to be "failed". This also prevents that the case of
	// already entered characeters changes.
	if (TCSLEN(lpCommand) == TCSLEN(lpBestMatch))
		return CR_FAILED;

	_sntprintf(lpBuffer, dwBufferLen, TEXT("%s"), lpBestMatch);
	free(lpBestMatch);

	return (fExactMatch == TRUE) ? CR_EXACT : CR_PARTLY;
}
