/**
* @file
* Configuration path and file handling.
*
* The functions provided in this file are used to find, open and read
* the configuration file. There are several occurences of dynamic memory
* allocations. Most of them are temporarily and will be deallocated at
* the end of the allocating functions. Some memory allocations are more
* persistent and the caller gains responsibility to free them. This is
* documented in the exported functions' descriptions.
*/

/*
 * Copyright (c) 2009, Felix Opatz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of  nor the names of its contributors may be used to
 *    endorse or promote products derived from this software without specific
 *    prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <assert.h>
#include <windows.h>
#include <tchar.h>

#include "AppLaunch.h"
#include "Config.h"
#include "ErrorHandling.h"
#include "StringUtil.h"

/****************************************************************************
* FILE LOCAL DEFINITIONS
*****************************************************************************/

/**
* Macro to mark a function exit as common jump label
*/
#define EXIT                ExitLabel:

/**
* Macro to set a return variable and leave the function by jumping to the
* common exit label.
* @param var Variable to set
* @param result Value to set
*/
#define LEAVE(var,result)   {(var) = (result); goto ExitLabel;}

/**
* Enumeration of keywords to provide keyword codes for parsing.
*/
enum Keywords
{
	KW_INVALID,            /** Invalid keyword       */
	KW_BACKGROUND_COLOR,   /** BackgroundColor       */
	KW_TEXT_COLOR,         /** TextColor             */
	KW_HOT_KEY,            /** HotKey                */
	KW_COMMAND,            /** Command               */
	KW_SPLASH_VISIBLE_TIME /** Splash screen timeout */
};

/**
* Results of attempt to insert a config entry into the config data structure
*/
enum InsertResults
{
	IR_OK,                 /** Insertion succeeded  */
	IR_INVALID_KEYWORD,    /** Keyword is invalid   */
	IR_MISSING_VALUE,      /** A value is missing   */
	IR_INVALID_VALUE,      /** The value is invalid */
	IR_NO_MEMORY           /** Out of memory        */
};

/**
* Entry for the lookup-table to translate a key description in textual form
* into a virtual key code usable as hotkey.
*/
struct VirtKeyTableEntry
{
	LPCTSTR lpString;      /** String with textual description      */
	UINT uCode;            /** Virtual key code assocatied with key */
};

/**
* Lookup-table to translate key descriptions into virtual key codes and vice
* versa.
*/
static struct VirtKeyTableEntry VirtKeyTable[] =
{
	{ TEXT("F1"),        VK_F1        },  { TEXT("F2"),        VK_F2        },  { TEXT("F3"),        VK_F3        },
	{ TEXT("F4"),        VK_F4        },  { TEXT("F5"),        VK_F5        },  { TEXT("F6"),        VK_F6        },
	{ TEXT("F7"),        VK_F7        },  { TEXT("F8"),        VK_F8        },  { TEXT("F9"),        VK_F9        },
	{ TEXT("F10"),       VK_F10       },  { TEXT("F11"),       VK_F11       },  { TEXT("F12"),       VK_F12       },

	{ TEXT("Tab"),       VK_TAB       },  { TEXT("Space"),     VK_SPACE     },  { TEXT("Escape"),    VK_ESCAPE    },
	{ TEXT("PageUp"),    VK_PRIOR     },  { TEXT("PageDown"),  VK_NEXT      },  { TEXT("End"),       VK_END       },
	{ TEXT("Home"),      VK_HOME      },  { TEXT("Insert"),    VK_INSERT    },  { TEXT("Delete"),    VK_DELETE    },

	{ TEXT("0"),         '0'          },  { TEXT("1"),         '1'          },  { TEXT("2"),         '2'          },
	{ TEXT("3"),         '3'          },  { TEXT("1"),         '4'          },  { TEXT("2"),         '5'          },
	{ TEXT("6"),         '6'          },  { TEXT("1"),         '7'          },  { TEXT("2"),         '8'          },
	{ TEXT("9"),         '9'          },  { TEXT("+"),         VK_ADD       },  { TEXT("-"),         VK_SUBTRACT  },
	{ TEXT("*"),         VK_MULTIPLY  },  { TEXT("/"),         VK_DIVIDE    },

	{ TEXT("A"),         'A'          },  { TEXT("B"),         'B'          },  { TEXT("C"),         'C'          },
	{ TEXT("D"),         'D'          },  { TEXT("E"),         'E'          },  { TEXT("F"),         'F'          },
	{ TEXT("G"),         'G'          },  { TEXT("H"),         'H'          },  { TEXT("I"),         'I'          },
	{ TEXT("J"),         'J'          },  { TEXT("K"),         'K'          },  { TEXT("L"),         'L'          },
	{ TEXT("M"),         'M'          },  { TEXT("N"),         'N'          },  { TEXT("O"),         'O'          },
	{ TEXT("P"),         'P'          },  { TEXT("Q"),         'Q'          },  { TEXT("R"),         'R'          },
	{ TEXT("S"),         'S'          },  { TEXT("T"),         'T'          },  { TEXT("U"),         'U'          },
	{ TEXT("V"),         'V'          },  { TEXT("W"),         'W'          },  { TEXT("X"),         'X'          },
	{ TEXT("Y"),         'Y'          },  { TEXT("Z"),         'Z'          },
};
/**
* Size of lookup-table
*/
#define VIRT_KEY_TABLE_SIZE (sizeof(VirtKeyTable)/sizeof(VirtKeyTable[0]))

/**
* Maximum length of command line
*/
#define MAX_CMDLINE_LEN 32768

/****************************************************************************
* FILE LOCAL FUNCTIONS
*****************************************************************************/

/**
* Initializes a CONFIG_DATA variable by setting the default values that
* may get overwritten or invalid markers for configurations that must be
* set later on.
* @param lpConfigData Pointer to configuration data
*/
static void InitConfigData(CONFIG_DATA *lpConfigData)
{
	// Default settings
	lpConfigData->fDataValid = TRUE;
	lpConfigData->crBackgroundColor = CONFIG_DEF_BACKGROUND_COLOR;
	lpConfigData->crTextColor = CONFIG_DEF_TEXT_COLOR;

	// Invalid markers
	lpConfigData->uHotKeyModifiers = INVALID_HOT_KEY_MODIFIER;
	lpConfigData->uHotKeyVirtKey = INVALID_HOT_KEY_VIRT_KEY;

	// Command entries
	lpConfigData->Commands = NULL;
	lpConfigData->dwCommandCount = 0;

	// Splash-Screen timeout
	lpConfigData->dwSplashVisibleTime = CONFIG_DEF_SPLASH_VISIBLE_TIME;
}

/**
* Cuts a trailing newline in the string. Actually it zero-terminates the
* string just at the first occuring newline.
* @param lpString String to cut (must be writable)
*/
static void CutNewline(LPTSTR lpString)
{
	while ((*lpString != '\0') && (*lpString != TEXT('\n')))
		lpString++;

	if (*lpString == TEXT('\n'))
		*lpString = '\0';
}

/**
* Returns TRUE if line is a comment (first non-empty character is #).
* @param lpString String to test
* @return TRUE if line is a comment
*/
static BOOL IsComment(LPCTSTR lpString)
{
	while ((*lpString != '\0') && (_istspace(*lpString) != 0))
		lpString++;

	return *lpString == TEXT('#');
}

/**
* Returns TRUE if line is empty. Note that lines containing only white space
* characters are considered as empty lines as well. This differs from other
* definitions of the term 'empty'.
* @param lpString String to test
* @return TRUE if string is empty
*/
static BOOL IsEmpty(LPCTSTR lpString)
{
	while ((*lpString != '\0') && (_istspace(*lpString) != 0))
		lpString++;

	return *lpString == '\0';
}

/**
* Splits a config line into keyword and value. Keyword is the part from the
* first non-white-space to the colon, not including the colon. The value begins
* with the first non-white-space after the colon and goes on to the end of the
* string. Trailing white space characters may occur in the value.
* The destination buffers MUST be large enough to receive the resulting strings
* or overflows will occur.
* @param lpString Configuration line to split (treated as read-only)
* @param lpKeyword Pointer to buffer that receives the keyword
* @param lpValue Pointer to buffer that receives the value
*/
static void GetKeyValuePair(LPCTSTR lpString, LPTSTR lpKeyword, LPTSTR lpValue)
{
	enum {START, IN_KEYWORD, AFTER_KEYWORD, IN_VALUE} iState = START;

	while (*lpString != '\0')
	{
		switch (iState)
		{
		case START:
			if (_istspace(*lpString) != 0)
				break;
			iState = IN_KEYWORD;
			continue;

		case IN_KEYWORD:
			if ((*lpString != TEXT(':')) && (_istspace(*lpString) == 0))
			{
				*lpKeyword = *lpString;
				lpKeyword++;
			}
			else
			{
				*lpKeyword = '\0';
				iState = AFTER_KEYWORD;
			}
			break;

		case AFTER_KEYWORD:
			if ((*lpString != TEXT(':')) && (_istspace(*lpString) == 0))
			{
				iState = IN_VALUE;
				continue;
			}
			break;

		case IN_VALUE:
			*lpValue = *lpString;
			lpValue++;
			break;
		}

		lpString++;
	}

	*lpKeyword = '\0';
	*lpValue = '\0';
}

/**
* Examines a string and returns a keyword enum value
* @param lpKeyword String to examine
* @return Enum value of keyword
*/
static enum Keywords GetKeywordCode(LPCTSTR lpKeyword)
{
	if (EQUALS(lpKeyword, TEXT("BackgroundColor")))   return KW_BACKGROUND_COLOR;
	if (EQUALS(lpKeyword, TEXT("TextColor")))         return KW_TEXT_COLOR;
	if (EQUALS(lpKeyword, TEXT("HotKey")))            return KW_HOT_KEY;
	if (EQUALS(lpKeyword, TEXT("Command")))           return KW_COMMAND;
	if (EQUALS(lpKeyword, TEXT("SplashVisibleTime"))) return KW_SPLASH_VISIBLE_TIME;

	return KW_INVALID;
}

/**
* Parses a RGB hex string (#xxxxxx) into a COLORREF structure. The result
* variable is only written to if the conversion was successful.
* @param lpColorString Hex string with RGB information
* @param lpColorResult Pointer to COLORREF structure receiving the result
* @return TRUE if conversion was successful
*/
static BOOL GetColorCode(LPCTSTR lpColorString, COLORREF *lpColorResult)
{
	TCHAR szColorComponent[3] = {'\0'};
	DWORD dwRed, dwGreen, dwBlue;
	TCHAR *lpEnd;

	// Marker for hex code
	if (*lpColorString != TEXT('#'))
		return FALSE;
	lpColorString++;

	// Remainder of string must be 6 characters
	if (TCSLEN(lpColorString) != 6)
		return FALSE;

	// RED component
	szColorComponent[0] = *(lpColorString++);
	szColorComponent[1] = *(lpColorString++);
	dwRed = _tcstol(szColorComponent, &lpEnd, 16);
	if (*lpEnd != '\0')
		return FALSE;
	
	// GREEN component
	szColorComponent[0] = *(lpColorString++);
	szColorComponent[1] = *(lpColorString++);
	dwGreen = _tcstol(szColorComponent, &lpEnd, 16);
	if (*lpEnd != '\0')
		return FALSE;
	
	// BLUE component
	szColorComponent[0] = *(lpColorString++);
	szColorComponent[1] = *(lpColorString++);
	dwBlue = _tcstol(szColorComponent, &lpEnd, 16);
	if (*lpEnd != '\0')
		return FALSE;
	
	// Construct COLORREF
	*lpColorResult = RGB(dwRed, dwGreen, dwBlue);
	return TRUE;
}

/**
* Tries to identify the token as key modifier and then adds (bitwise OR) it
* to the already existing modifiers.
* @param lpToken Token that may contain a key modifier
* @param lpModifiers Pointer to an UINT to receive modifiers
* @return TRUE if token was a valid modifier
*/
static BOOL AddModifier(LPCTSTR lpToken, UINT *lpModifiers)
{
	if (EQUALS(lpToken, TEXT("Shift")))
	{
		*lpModifiers |= MOD_SHIFT;
		return TRUE;
	}

	if (EQUALS(lpToken, TEXT("Ctrl")))
	{
		*lpModifiers |= MOD_CONTROL;
		return TRUE;
	}

	if (EQUALS(lpToken, TEXT("Alt")))
	{
		*lpModifiers |= MOD_ALT;
		return TRUE;
	}

	if (EQUALS(lpToken, TEXT("Win")))
	{
		*lpModifiers |= MOD_WIN;
		return TRUE;
	}

	return FALSE;
}

/**
* Tries to identify the token as virtual key and then writes its code to the
* result parameter.
* @param lpToken Token that may contain a virtual key
* @param lpKeyCode Pointer to an UINT to receive the virtual key code
* @return TRUE if token was a valid virtual key
*/
static BOOL AddVirtualKey(LPCTSTR lpToken, UINT *lpKeyCode)
{
	int iTablePos;

	for (iTablePos = 0; iTablePos < VIRT_KEY_TABLE_SIZE; iTablePos++)
		if (EQUALS(lpToken, VirtKeyTable[iTablePos].lpString))
		{
			*lpKeyCode = VirtKeyTable[iTablePos].uCode;
			return TRUE;
		}

	return FALSE;
}

/**
* Parses a hot key definition and inserts the key modifiers and the virtual key
* code. The result variables are only written to if the conversion was
* successful.
* @param lpKeyDefinition Key definition to parse
* @param lpKeyModifiers Pointer to an UINT to receive key modifiers
* @param lpKeyModifiers Pointer to an UINT to receive the virtual key code
* @return TRUE if conversion was successful
*/
static BOOL GetHotKey(LPCTSTR lpKeyDefinition, UINT *lpKeyModifiers, UINT *lpVirtualKeyCode)
{
	UINT uKeyModifiers = 0;
	UINT uVirtKeyCode = 0;
	TCHAR szCurrToken[MAX_CONFIG_LINE_LEN + 1];
	TCHAR *lpNextChar;
	BOOL fSucceeded = FALSE;

	lpNextChar = szCurrToken;
	while (*lpKeyDefinition != '\0')
	{
		if (_istspace(*lpKeyDefinition) == 0)
		{
			// Not a space
			*lpNextChar = *lpKeyDefinition;
			lpNextChar++;
			lpKeyDefinition++;
		}
		else
		{
			// A space
			*lpNextChar = '\0';
			lpNextChar = szCurrToken;
			lpKeyDefinition++;

			// Try to add token as modifier
			if (AddModifier(szCurrToken, &uKeyModifiers) == TRUE)
				continue;

			// Try to add token as virtual key
			if (AddVirtualKey(szCurrToken, &uVirtKeyCode) == TRUE)
				continue;

			// This token is not usable
			return FALSE;
		}
	}

	*lpNextChar = '\0';

	// Try to add token as modifier
	if (AddModifier(szCurrToken, &uKeyModifiers) == TRUE)
		fSucceeded = TRUE;

	// Try to add token as virtual key
	if (AddVirtualKey(szCurrToken, &uVirtKeyCode) == TRUE)
		fSucceeded = TRUE;

	// Adding last token successful?
	if (fSucceeded == TRUE)
	{
		*lpKeyModifiers = uKeyModifiers;
		*lpVirtualKeyCode = uVirtKeyCode;
	}

	return fSucceeded;
}

/**
* Parses a command entry. 
* @param lpCommand String containing command entry to parse
* @param szName Pointer to buffer that receives the command's name
* @param szPath Pointer to buffer that receives the command's path
* @param szWorkingDir Pointer to buffer that receives the working directory
* @param szOptions Pointer to buffer that receives optional arguments
* @return TRUE if command was successfully parsed into components
*/
static BOOL GetCommand(LPCTSTR lpCommand, TCHAR szName[MAX_PATH],
		       TCHAR szPath[MAX_PATH], TCHAR szWorkingDir[MAX_PATH],
		       TCHAR szOptions[MAX_CMDLINE_LEN])
{
	enum {IN_NAME, AFTER_NAME, IN_PATH, IN_QUOTED_PATH, AFTER_PATH, IN_DIR,
		IN_QUOTED_DIR, AFTER_DIR, DONE} iState = IN_NAME;
	TCHAR *lpNextChar;

	szName[0] = '\0';
	szPath[0] = '\0';
	szWorkingDir[0] = '\0';
	szOptions[0] = '\0';
	lpNextChar = szName;
	while ((*lpCommand != '\0') && (iState != DONE))
	{
		switch (iState)
		{
		case IN_NAME:
			if (_istspace(*lpCommand) != 0)
			{
				*lpNextChar = '\0';
				iState = AFTER_NAME;
				continue;
			}
			else
			{
				*lpNextChar = *lpCommand;
				lpNextChar++;
				lpCommand++;
			}
			break;

		case AFTER_NAME:
			if (_istspace(*lpCommand) != 0)
			{
				lpCommand++;
			}
			else
			{
				iState = IN_PATH;
				lpNextChar = szPath;
				continue;
			}
			break;

		case IN_PATH:
			if (*lpCommand == TEXT('"'))
			{
				lpCommand++;
				iState = IN_QUOTED_PATH;
				break;
			}

			if (_istspace(*lpCommand) != 0)
			{
				*lpNextChar = '\0';
				iState = AFTER_PATH;
				continue;
			}

			*lpNextChar = *lpCommand;
			lpNextChar++;
			lpCommand++;
			break;

		case IN_QUOTED_PATH:
			if (*lpCommand == TEXT('"'))
			{
				*lpNextChar = '\0';
				lpCommand++;
				iState = AFTER_PATH;
				continue;
			}

			*lpNextChar = *lpCommand;
			lpNextChar++;
			lpCommand++;
			break;

		case AFTER_PATH:
			if (_istspace(*lpCommand) != 0)
			{
				lpCommand++;
			}
			else
			{
				iState = IN_DIR;
				lpNextChar = szWorkingDir;
				continue;
			}
			break;

		case IN_DIR:
			if (*lpCommand == TEXT('"'))
			{
				lpCommand++;
				iState = IN_QUOTED_DIR;
				break;
			}

			if (_istspace(*lpCommand) != 0)
			{
				*lpNextChar = '\0';
				iState = AFTER_DIR;
				continue;
			}

			*lpNextChar = *lpCommand;
			lpNextChar++;
			lpCommand++;
			break;

		case IN_QUOTED_DIR:
			if (*lpCommand == TEXT('"'))
			{
				*lpNextChar = '\0';
				lpCommand++;
				iState = AFTER_DIR;
				continue;
			}

			*lpNextChar = *lpCommand;
			lpNextChar++;
			lpCommand++;
			break;

		case AFTER_DIR:
			if (_istspace(*lpCommand) != 0)
				lpCommand++;
			else
				iState = DONE;
			break;
		}
		
	}
	*lpNextChar = '\0';

	/* If either name or path isn't present the command was invalid */
	if (TCSLEN(szName) == 0 || TCSLEN(szPath) == 0)
		return FALSE;

	if (TCSLEN(lpCommand) > 0)
		_tcscpy(szOptions, lpCommand);

	return TRUE;
}

/**
* Parses and inserts a command entry. The result variables are only changed
* if the command entry is valid, except that lpEntries may be NULL if reallocation
* of a large enough piece of memory failed.
* @param lpCommand String containing command entry to parse
* @param lpEntries Pointer to array in which the entry will be inserted
* @param lpCount Pointer to the entry counter
* @return TRUE if action succeeded
*/
static BOOL InsertCommand(COMMAND_ENTRY **lpEntries, DWORD *lpCount, int iCmdType, LPCTSTR lpName,
			  LPCTSTR lpPath, LPCTSTR lpWorkingDir, LPCTSTR lpOptions)
{
	BOOL fSucceeded = FALSE;
	COMMAND_ENTRY NewEntry;
	int iInsertPos;
	int iShiftPos;

	NewEntry.iCmdType = iCmdType;

	// Initialize all pointers, the optional ones are only overwritten
	// if the target is valid
	NewEntry.lpName = NULL;
	NewEntry.lpPath = NULL;
	NewEntry.lpWorkingDir = NULL;
	NewEntry.lpOptions = NULL;

	NewEntry.lpName = _tcsdup(lpName);
	if (NewEntry.lpName == NULL)
		LEAVE(fSucceeded, FALSE);

	if (lpPath)
	{
		NewEntry.lpPath = _tcsdup(lpPath);
		if (NewEntry.lpPath == NULL)
			LEAVE(fSucceeded, FALSE);
	}
	
	if (lpWorkingDir && (TCSLEN(lpWorkingDir) > 0) && (_tcscmp(lpWorkingDir, TEXT("-")) != 0))
	{
		NewEntry.lpWorkingDir = _tcsdup(lpWorkingDir);
		if (NewEntry.lpWorkingDir == NULL)
			LEAVE(fSucceeded, FALSE);
	}

	if (lpOptions && TCSLEN(lpOptions) > 0)
	{
		NewEntry.lpOptions = _tcsdup(lpOptions);
		if (NewEntry.lpOptions == NULL)
			LEAVE(fSucceeded, FALSE);
	}

	// Allocate more memory to insert new entry
	*lpEntries = realloc(*lpEntries, (*lpCount + 1) * sizeof(**lpEntries));

	// Out of memory
	if (*lpEntries == NULL)
		LEAVE(fSucceeded, FALSE);

	(*lpCount)++;

	// Find position to insert new entry
	for (iInsertPos = 0; iInsertPos < (int)*lpCount - 1; iInsertPos++)
		if (CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, NewEntry.lpName, -1,
			(*lpEntries)[iInsertPos].lpName, -1) == CSTR_LESS_THAN)
		{
			break;
		}

	// Shift entries to the end (if needed)
	for (iShiftPos = *lpCount - 1; iShiftPos > iInsertPos; iShiftPos--)
		(*lpEntries)[iShiftPos] = (*lpEntries)[iShiftPos - 1];

	// Insert new entry
	(*lpEntries)[iInsertPos] = NewEntry;

	LEAVE(fSucceeded, TRUE);

EXIT
	// If allocating enough memory for all entries the action will not be
	// completed and already allocated memory will be freed.
	if (fSucceeded == FALSE)
	{
		if (NewEntry.lpName != NULL)
		{
			free((void*)NewEntry.lpName);
			NewEntry.lpName = NULL;
		}

		if (NewEntry.lpPath != NULL)
		{
			free((void*)NewEntry.lpPath);
			NewEntry.lpPath = NULL;
		}

		if (NewEntry.lpWorkingDir != NULL)
		{
			free((void*)NewEntry.lpWorkingDir);
			NewEntry.lpWorkingDir = NULL;
		}

		if (NewEntry.lpOptions != NULL)
		{
			free((void*)NewEntry.lpOptions);
			NewEntry.lpOptions = NULL;
		}
	}

	return fSucceeded;
}

/**
* Frees the memory allocated for strings in each command entry, but not the
* command list itself.
* @param lpEntries Pointer to an array of command entries
* @param dwCount Number of entries in the array
*/
static void FreeCommandList(COMMAND_ENTRY *lpEntries, DWORD dwCount)
{
	while (dwCount > 0)
	{
		if (lpEntries->lpName != NULL)
		{
			free((void*)lpEntries->lpName);
			lpEntries->lpName = NULL;
		}

		if (lpEntries->lpPath != NULL)
		{
			free((void*)lpEntries->lpPath);
			lpEntries->lpPath = NULL;
		}

		if (lpEntries->lpWorkingDir != NULL)
		{
			free((void*)lpEntries->lpWorkingDir);
			lpEntries->lpWorkingDir = NULL;
		}

		if (lpEntries->lpOptions != NULL)
		{
			free((void*)lpEntries->lpOptions);
			lpEntries->lpOptions = NULL;
		}

		lpEntries++;
		dwCount--;
	}
}

/**
* Inserts a given piece of config data into the data structure.
* @param lpKeyword Keyword of config data piece
* @param lpValue Value of config data piece
* @param lpConfigData Pointer to structure to store data in
* @return Error code indicating the success of the insertion
*/
static enum InsertResults InsertConfigData(LPCTSTR lpKeyword, LPCTSTR lpValue, CONFIG_DATA *lpConfigData)
{
	TCHAR *pEnd;
	DWORD dwTimeout;
	BOOL fResult;
	TCHAR szName[MAX_PATH];
	TCHAR szPath[MAX_PATH];
	TCHAR szWorkingDir[MAX_PATH];
	TCHAR szOptions[MAX_CMDLINE_LEN];

	// Test if value is present
	if (IsEmpty(lpValue) == TRUE)
		return IR_MISSING_VALUE;

	switch (GetKeywordCode(lpKeyword))
	{
		// Invalid keyword passed
		default:
		case KW_INVALID:
			return IR_INVALID_KEYWORD;

		case KW_BACKGROUND_COLOR:
			if (GetColorCode(lpValue, &lpConfigData->crBackgroundColor) == FALSE)
				return IR_INVALID_VALUE;
			break;

		case KW_TEXT_COLOR:
			if (GetColorCode(lpValue, &lpConfigData->crTextColor) == FALSE)
				return IR_INVALID_VALUE;
			break;

		case KW_HOT_KEY:
			if (GetHotKey(lpValue, &lpConfigData->uHotKeyModifiers, &lpConfigData->uHotKeyVirtKey) == FALSE)
				return IR_INVALID_VALUE;
			break;

		case KW_COMMAND:
			fResult = GetCommand(lpValue, szName, szPath, szWorkingDir, szOptions);
			if (fResult == FALSE)
				return IR_INVALID_VALUE;

			fResult = InsertCommand(&lpConfigData->Commands, &lpConfigData->dwCommandCount,
				CMD_TYPE_EXTERNAL, szName, szPath, szWorkingDir, szOptions);
			if (fResult == FALSE)
				return IR_NO_MEMORY;
			break;

		case KW_SPLASH_VISIBLE_TIME:
			dwTimeout = _tcstoul(lpValue, &pEnd, 0);
			if (*pEnd != TEXT('\0'))
				return IR_INVALID_VALUE;
			lpConfigData->dwSplashVisibleTime = dwTimeout;
			break;
	}

	return IR_OK;
}

/****************************************************************************
* GLOBAL FUNCTIONS
*****************************************************************************/

/**
* Generates the path where the configuration file should be located by
* evaluating the environment variable APP_DATA_ENV_NAME.
* @param lpPathBuffer Pointer to a buffer that receives the path
* @param dwBufferLen Length of buffer in characters
* @return On success the return value is nonzero
*/
BOOL GetConfigPath(LPTSTR lpBuffer, DWORD nSize)
{
	TCHAR *lpAppDataBuffer = NULL;
	DWORD dwEnvResult;
	DWORD dwAppDataCch;
	int iResultLen;
	BOOL fSucceeded = FALSE;

	// Query needed buffer size (in characters)
	dwAppDataCch = GetEnvironmentVariable(CONFIG_APPDATA_ENV_NAME, NULL, 0);

	// Call failed, e.g. because variable is not set
	if (dwAppDataCch == 0)
		LEAVE(fSucceeded, FALSE);

	// Allocate enough memory to receive environment value
	lpAppDataBuffer = malloc(dwAppDataCch * sizeof(TCHAR));

	// Assure that memory allocation succeeded
	if (lpAppDataBuffer == NULL)
		LEAVE(fSucceeded, FALSE);

	// Query environment variable
	dwEnvResult = GetEnvironmentVariable(CONFIG_APPDATA_ENV_NAME, lpAppDataBuffer, dwAppDataCch);

	// If everything is fine the number of stored characters should be the
	// number previously announced, excluding the terminating zero character.
	assert(dwEnvResult == dwAppDataCch - 1);

	// Combine the path to the user's application data, the sub directory
	// for this application, and the name of the configuration file.
	iResultLen = _sntprintf(lpBuffer, nSize, TEXT("%s\\%s\\%s"),
		lpAppDataBuffer, CONFIG_APP_SUBDIR_NAME, CONFIG_APP_FILE_NAME);

	// Buffer too small, or string fits exactly but is not terminated and
	// thus also unusable
	if ((iResultLen < 0) || (iResultLen == (int)nSize))
		LEAVE(fSucceeded, FALSE);

	LEAVE(fSucceeded, TRUE);

EXIT
	// Deallocate memory if needed
	if (lpAppDataBuffer != NULL)
	{
		free(lpAppDataBuffer);
		lpAppDataBuffer = NULL;
	}

	return fSucceeded;
}

/**
* Loads the configuration from the specified file. This function allocates
* memory for the configuration data. A call to FreeConfig must be made to
* prevent memory leaks.
* @param lpConfigPath Path to configuration file
* @param lpConfigData Pointer to data structure to store configuration in
* @return On success the return value is nonzero
*/
BOOL LoadConfig(LPCTSTR lpConfigPath, CONFIG_DATA *lpConfigData)
{
	BOOL fSucceeded = FALSE;
	FILE *lpFile = NULL;
	TCHAR szCurrLine[MAX_CONFIG_LINE_LEN + 1];
	TCHAR szKeyword[MAX_CONFIG_LINE_LEN + 1];
	TCHAR szValue[MAX_CONFIG_LINE_LEN + 1];
	TCHAR szNotepadPath[MAX_PATH];
	DWORD dwCurrLineNo = 0;
	int iInsertResult;

	// Initialize data structure and set defaults
	InitConfigData(lpConfigData);

	lpFile = _tfopen(lpConfigPath, TEXT("r"));

	// Provide user feedback if opening file failed
	if (lpFile == NULL)
	{
		ReportError(ERROR_OPENING_CONFIG_FILE, lpConfigPath, errno);
		LEAVE(fSucceeded, FALSE);
	}

	while (_fgetts(szCurrLine, sizeof(szCurrLine), lpFile) != NULL)
	{
		// szCurrLine now contains a zero-terminated string which either is a
		// complete line (line length is smaller than buffer size) or a fragment
		// (which is considered as user-caused problem because of too long lines).
		// If there is a trailing newline it is stored in the string, the last line
		// may as well be unfinished (we can deal with that).

		CutNewline(szCurrLine);
		dwCurrLineNo++;

		// Skip if comment
		if (IsComment(szCurrLine) == TRUE)
			continue;

		// Skip if line is empty
		if (IsEmpty(szCurrLine) == TRUE)
			continue;

		GetKeyValuePair(szCurrLine, szKeyword, szValue);

		// Try to insert config data and report error if failed
		iInsertResult = InsertConfigData(szKeyword, szValue, lpConfigData);
		switch (iInsertResult)
		{
		case IR_INVALID_KEYWORD:
			ReportError(ERROR_INVALID_KEYWORD, dwCurrLineNo, szKeyword);
			LEAVE(fSucceeded, FALSE);

		case IR_MISSING_VALUE:
			ReportError(ERROR_MISSING_VALUE, dwCurrLineNo, szKeyword);
			LEAVE(fSucceeded, FALSE);

		case IR_INVALID_VALUE:
			ReportError(ERROR_INVALID_VALUE, dwCurrLineNo, szValue);
			LEAVE(fSucceeded, FALSE);

		case IR_NO_MEMORY:
			ReportError(ERROR_NO_MEMORY, dwCurrLineNo, szValue);
			LEAVE(fSucceeded, FALSE);
		}	
	}

	// Add pseudo-commands
	GetWindowsDirectory(szNotepadPath, sizeof(szNotepadPath));
	_tcscat(szNotepadPath, TEXT("\\Notepad.exe"));
	InsertCommand(&lpConfigData->Commands, &lpConfigData->dwCommandCount, CMD_TYPE_EXTERNAL,
		TEXT(":config"), szNotepadPath, NULL, lpConfigPath);
	InsertCommand(&lpConfigData->Commands, &lpConfigData->dwCommandCount, CMD_TYPE_INTERNAL_QUIT,
		TEXT(":quit"), NULL, NULL, NULL);
	InsertCommand(&lpConfigData->Commands, &lpConfigData->dwCommandCount, CMD_TYPE_INTERNAL_RELOAD,
		TEXT(":reload"), NULL, NULL, NULL);

	LEAVE(fSucceeded, TRUE);

EXIT
	// Close file if opening was successful
	if (lpFile != NULL)
	{
		fclose(lpFile);
		lpFile = NULL;
	}

	// Free command list if necessary
	if ((fSucceeded == FALSE) && (lpConfigData->dwCommandCount > 0))
	{
		FreeCommandList(lpConfigData->Commands, lpConfigData->dwCommandCount);
		free(lpConfigData->Commands);
		lpConfigData->Commands = NULL;
		lpConfigData->dwCommandCount = 0;
	}

	return fSucceeded;
}

/**
* Deallocates the memory allocated using LoadConfig. The content of the
* configuration is undefined after the functions returns.
* @param lpConfigData Pointer to data structure to free
*/
void FreeConfig(CONFIG_DATA *lpConfigData)
{
	// Do nothing if data is not valid (e.g. already freed once)
	if (lpConfigData->fDataValid == FALSE)
		return;

	// Free command list if necessary
	if (lpConfigData->dwCommandCount > 0)
	{
		FreeCommandList(lpConfigData->Commands, lpConfigData->dwCommandCount);
		free(lpConfigData->Commands);
		lpConfigData->Commands = NULL;
		lpConfigData->dwCommandCount = 0;
	}

	lpConfigData->fDataValid = FALSE;
}

/**
* Returns if hot key is configured
* @return TRUE if hot key is configured
*/
BOOL IsHotKeyValid(UINT uModifiers, UINT uKeyCode)
{
	if ((uModifiers == INVALID_HOT_KEY_MODIFIER) ||
		(uKeyCode == INVALID_HOT_KEY_VIRT_KEY))
	{
		return FALSE;
	}

	return TRUE;
}

/**
* Constructs a string representing the current configured hot key.
* @param lpBuffer Pointer to buffer to receive string
* @param dwBufferLength Size of buffer in characters
* @param uModifiers Modifiers of HotKey
* @param uKeyCode Virtual key code of HotKey
* @return TRUE if string construction was successful
*/
BOOL GetHotKeyAsString(LPTSTR lpBuffer, DWORD dwBufferLength, UINT uModifiers, UINT uKeyCode)
{
	DWORD dwCharsWritten = 0;
	LPCTSTR lpShiftKey = TEXT("Shift + ");
	LPCTSTR lpCtrlKey = TEXT("Ctrl + ");
	LPCTSTR lpAltKey = TEXT("Alt + ");
	LPCTSTR lpWinKey = TEXT("Win + ");
	int iTablePos;

	if (!IsHotKeyValid(uModifiers, uKeyCode))
		return FALSE;

	// Write the zero termination, and to make the following code simpler the
	// 1 character for the termination is subtracted from the overall length.
	*lpBuffer = '\0';
	dwBufferLength--;

	if (uModifiers & MOD_SHIFT)
	{
		if (dwCharsWritten + TCSLEN(lpShiftKey) <= dwBufferLength)
		{		
			_tcscat(lpBuffer, lpShiftKey);
			dwCharsWritten += TCSLEN(lpShiftKey);
		}
		else
		{
			return FALSE;
		}
	}

	if (uModifiers & MOD_CONTROL)
	{
		if (dwCharsWritten + TCSLEN(lpCtrlKey) <= dwBufferLength)
		{		
			_tcscat(lpBuffer, lpCtrlKey);
			dwCharsWritten += TCSLEN(lpCtrlKey);
		}
		else
		{
			return FALSE;
		}
	}

	if (uModifiers & MOD_ALT)
	{
		if (dwCharsWritten + TCSLEN(lpAltKey) <= dwBufferLength)
		{		
			_tcscat(lpBuffer, lpAltKey);
			dwCharsWritten += TCSLEN(lpAltKey);
		}
		else
		{
			return FALSE;
		}
	}

	if (uModifiers & MOD_WIN)
	{
		if (dwCharsWritten + TCSLEN(lpWinKey) <= dwBufferLength)
		{		
			_tcscat(lpBuffer, lpWinKey);
			dwCharsWritten += TCSLEN(lpWinKey);
		}
		else
		{
			return FALSE;
		}
	}

	// Attach the virtual key and return with TRUE if a key was found
	for (iTablePos = 0; iTablePos < VIRT_KEY_TABLE_SIZE; iTablePos++)
	{
		if (VirtKeyTable[iTablePos].uCode == uKeyCode)
		{
			if (dwCharsWritten + TCSLEN(VirtKeyTable[iTablePos].lpString) <= dwBufferLength)
			{		
				_tcscat(lpBuffer, VirtKeyTable[iTablePos].lpString);
				dwCharsWritten += TCSLEN(VirtKeyTable[iTablePos].lpString);
				return TRUE;
			}
			else
			{
				return FALSE;
			}
		}
	}

	// Not virtual key was set, must be a wrong configuration!
	return FALSE;
}
