/**
* @file
* Interface for configuration file handling.
*/

/*
 * Copyright (c) 2009, Felix Opatz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of  nor the names of its contributors may be used to
 *    endorse or promote products derived from this software without specific
 *    prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef CONFIG_H
#define CONFIG_H

/**
* Data set of a single command entry
*/
typedef struct
{
	int iCmdType;                 /** Type of command                     */
	LPCTSTR lpName;               /** Name of command (entered text)      */
	LPCTSTR lpPath;               /** Path of executable                  */
	LPCTSTR lpWorkingDir;         /** Working dir to change to            */
	LPCTSTR lpOptions;            /** Options to be passed to program     */
} COMMAND_ENTRY;

#define CMD_TYPE_INTERNAL_QUIT    1
#define CMD_TYPE_INTERNAL_RELOAD  2
#define CMD_TYPE_EXTERNAL         3

/**
* Data set containing all configurable settings.
*/
typedef struct
{
	BOOL fDataValid;              /** Valid flag for complete data set           */
	COLORREF crBackgroundColor;   /** Color of window background                 */
	COLORREF crTextColor;         /** Color of text                              */
	UINT uHotKeyModifiers;        /** Modifiers of hot key (MOD_ALT etc.)        */
	UINT uHotKeyVirtKey;          /** Virtual key code of hot key                */
	COMMAND_ENTRY *Commands;      /** Commands that can be entered               */
	DWORD dwCommandCount;         /** Length of array Commands                   */
	DWORD dwSplashVisibleTime;    /** Time Splash-Screen is visible (in seconds) */
} CONFIG_DATA;

/**
* Invalid marker for the key modifier
*/
#define INVALID_HOT_KEY_MODIFIER UINT_MAX

/**
* Invalid marker for the virtual key code
*/
#define INVALID_HOT_KEY_VIRT_KEY UINT_MAX

/**
* Default setting for background color
*/
#define CONFIG_DEF_BACKGROUND_COLOR RGB(0x88, 0x00, 0x88)

/**
* Default setting for text color
*/
#define CONFIG_DEF_TEXT_COLOR RGB(0xFF, 0xFF, 0xFF)

/**
* Default setting for splash screen timeout (in seconds)
*/
#define CONFIG_DEF_SPLASH_VISIBLE_TIME 5

/**
* Environment variable that holds the path where the user stores the
* data needed by applications (configuration etc.).
*/
#define CONFIG_APPDATA_ENV_NAME TEXT("AppData")

/**
* Name of the subdirectory in which the configuration data of AppLaunch
* actually is stored.
*/
#define CONFIG_APP_SUBDIR_NAME APP_NAME

/**
* Name of the configuration file.
*/
#define CONFIG_APP_FILE_NAME (APP_NAME TEXT(".conf"))

/**
* Maximum length of a line from the configuration file (including newline).
*/
#define MAX_CONFIG_LINE_LEN 4095

BOOL GetConfigPath(LPTSTR lpBuffer, DWORD nSize);
BOOL LoadConfig(LPCTSTR lpConfigPath, CONFIG_DATA *lpConfigData);
void FreeConfig(CONFIG_DATA *lpConfigData);
BOOL GetHotKeyAsString(LPTSTR lpBuffer, DWORD dwBufferLength, UINT uModifiers, UINT uKeyCode);
BOOL IsHotKeyValid(UINT uModifiers, UINT uKeyCode);

#endif
