/**
* @file
* Error Handling.
*
* The functions from this file are used to report errors to the user by
* constructing an error message and showing it using a message box.
*/

/*
 * Copyright (c) 2009, Felix Opatz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of  nor the names of its contributors may be used to
 *    endorse or promote products derived from this software without specific
 *    prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <assert.h>
#include <stdarg.h>
#include <errno.h>

#include <tchar.h>
#include <windows.h>

#include "AppLaunch.h"
#include "AppLaunchRes.h"
#include "ErrorHandling.h"
#include "StringUtil.h"

/**
* Maximum length of error descriptions.
*/
#define MAX_ERROR_DESCRIPTION_LEN 1024

/**
* Gets an textual error description from a GetLastError return value
* @param dwLastError Error value returned by GetLastError
* @return Pointer to textual description
*/
static LPCTSTR GetErrorDescriptionFromLastError(DWORD dwLastError)
{
	static TCHAR szErrorDescription[MAX_ERROR_DESCRIPTION_LEN];
	DWORD dwResult;

	dwResult = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, 0, dwLastError, 0,
		szErrorDescription, MAX_ERROR_DESCRIPTION_LEN, NULL);

	if (dwResult == 0)
	{
		assert(0);
		return TEXT("Invalid error code passed");
	}

	return szErrorDescription;
}

/**
* Reports an error from ErrorCodes to the user. The variable argument
* list allows the caller to pass detailed information about the error
* condition. The function "knows" which arguments are passed (as long as
* the caller plays fair) and pops them from the stack. Be sure that the
* arguments passed fit to the implicit format definition in form of the
* error code.
* @param nErrorCode Code of error to report
*/
void ReportError(const enum ErrorCodes nErrorCode, ...)
{
	va_list arg_ptr;
	TCHAR szCaption[] = APP_NAME;
	LPTSTR lpMessageBuffer = NULL;
	DWORD dwMessageCch;
	int iResultLen;
	LPCTSTR lpFormatString;

	LPCTSTR lpPath;
	LPCTSTR lpErrorDescription;
	int iLineNumber;
	LPCTSTR lpKeyword;
	LPCTSTR lpInvalidValue;
	LPCTSTR lpProgramPath;

	va_start(arg_ptr, nErrorCode);

	switch (nErrorCode)
	{
	case ERROR_APP_ALREADY_STARTED:
		// Load format string from resources
		LoadString(g_hInstance, IDS_ERROR_APP_ALREADY_STARTED, (LPTSTR)&lpFormatString, 0);

		// Simple message, doesn't have arguments
		lpMessageBuffer = _tcsdup(lpFormatString);

		break;

	case ERROR_OPENING_CONFIG_FILE:
		// Load format string from resources
		LoadString(g_hInstance, IDS_ERROR_OPENING_CONFIG_FILE, (LPTSTR)&lpFormatString, 0);
		
		// Harvest arguments to include in message
		lpPath = va_arg(arg_ptr, LPTSTR);
		lpErrorDescription = _tcserror(va_arg(arg_ptr, int));

		// Calculate needed memory and allocate it
		dwMessageCch = TCSLEN(lpFormatString) + TCSLEN(lpPath) + TCSLEN(lpErrorDescription);
		lpMessageBuffer = malloc(dwMessageCch * sizeof(TCHAR));

		// Assure that the memory could be allocated
		if (lpMessageBuffer == NULL)
			break;

		// Assemble message from parts
		iResultLen = _sntprintf(lpMessageBuffer, dwMessageCch - 1, lpFormatString,
			lpPath, lpErrorDescription);
		lpMessageBuffer[iResultLen] = '\0';

		break;

	case ERROR_INVALID_KEYWORD:
		// Load format string from resources
		LoadString(g_hInstance, IDS_ERROR_INVALID_KEYWORD, (LPTSTR)&lpFormatString, 0);
		
		// Harvest arguments to include in message
		iLineNumber = va_arg(arg_ptr, int);
		lpKeyword = va_arg(arg_ptr, LPTSTR);

		// Calculate needed memory and allocate it
		dwMessageCch = TCSLEN(lpFormatString) + 10 + TCSLEN(lpKeyword);
		lpMessageBuffer = malloc(dwMessageCch * sizeof(TCHAR));

		// Assure that the memory could be allocated
		if (lpMessageBuffer == NULL)
			break;

		// Assemble message from parts
		iResultLen = _sntprintf(lpMessageBuffer, dwMessageCch - 1, lpFormatString,
			iLineNumber, lpKeyword);
		lpMessageBuffer[iResultLen] = '\0';

		break;

	case ERROR_MISSING_VALUE:
		// Load format string from resources
		LoadString(g_hInstance, IDS_ERROR_MISSING_VALUE, (LPTSTR)&lpFormatString, 0);
		
		// Harvest arguments to include in message
		iLineNumber = va_arg(arg_ptr, int);
		lpKeyword = va_arg(arg_ptr, LPTSTR);

		// Calculate needed memory and allocate it
		dwMessageCch = TCSLEN(lpFormatString) + 10 + TCSLEN(lpKeyword);
		lpMessageBuffer = malloc(dwMessageCch * sizeof(TCHAR));

		// Assure that the memory could be allocated
		if (lpMessageBuffer == NULL)
			break;

		// Assemble message from parts
		iResultLen = _sntprintf(lpMessageBuffer, dwMessageCch - 1, lpFormatString,
			iLineNumber, lpKeyword);
		lpMessageBuffer[iResultLen] = '\0';

		break;

	case ERROR_INVALID_VALUE:
		// Load format string from resources
		LoadString(g_hInstance, IDS_ERROR_INVALID_VALUE, (LPTSTR)&lpFormatString, 0);
		
		// Harvest arguments to include in message
		iLineNumber = va_arg(arg_ptr, int);
		lpInvalidValue = va_arg(arg_ptr, LPTSTR);

		// Calculate needed memory and allocate it
		dwMessageCch = TCSLEN(lpFormatString) + 10 + TCSLEN(lpInvalidValue);
		lpMessageBuffer = malloc(dwMessageCch * sizeof(TCHAR));

		// Assure that the memory could be allocated
		if (lpMessageBuffer == NULL)
			break;

		// Assemble message from parts
		iResultLen = _sntprintf(lpMessageBuffer, dwMessageCch - 1, lpFormatString,
			iLineNumber, lpInvalidValue);
		lpMessageBuffer[iResultLen] = '\0';

		break;

	case ERROR_INVALID_LINE:
		// Load format string from resources
		LoadString(g_hInstance, IDS_ERROR_INVALID_LINE, (LPTSTR)&lpFormatString, 0);
		
		// Harvest arguments to include in message
		iLineNumber = va_arg(arg_ptr, int);

		// Calculate needed memory and allocate it
		dwMessageCch = TCSLEN(lpFormatString) + 10;
		lpMessageBuffer = malloc(dwMessageCch * sizeof(TCHAR));

		// Assure that the memory could be allocated
		if (lpMessageBuffer == NULL)
			break;

		// Assemble message from parts
		iResultLen = _sntprintf(lpMessageBuffer, dwMessageCch - 1, lpFormatString, iLineNumber);
		lpMessageBuffer[iResultLen] = '\0';

		break;

	case ERROR_CREATING_PROCESS:
		// Load format string from resources
		LoadString(g_hInstance, IDS_ERROR_CREATING_PROCESS, (LPTSTR)&lpFormatString, 0);
		
		// Harvest arguments to include in message
		lpProgramPath = va_arg(arg_ptr, LPTSTR);
		lpErrorDescription = GetErrorDescriptionFromLastError(va_arg(arg_ptr, DWORD));

		// Calculate needed memory and allocate it
		dwMessageCch = TCSLEN(lpFormatString) + TCSLEN(lpProgramPath) + TCSLEN(lpErrorDescription);
		lpMessageBuffer = malloc(dwMessageCch * sizeof(TCHAR));

		// Assure that the memory could be allocated
		if (lpMessageBuffer == NULL)
			break;

		// Assemble message from parts
		iResultLen = _sntprintf(lpMessageBuffer, dwMessageCch - 1, lpFormatString,
			lpProgramPath, lpErrorDescription);
		lpMessageBuffer[iResultLen] = '\0';

		break;

	case ERROR_NO_MEMORY:
		// Load format string from resources
		LoadString(g_hInstance, IDS_ERROR_NO_MEMORY, (LPTSTR)&lpFormatString, 0);

		// Simple message, doesn't have arguments
		lpMessageBuffer = _tcsdup(lpFormatString);
		break;

	case ERROR_REGISTER_HOTKEY_FAILED:
		// Load format string from resources
		LoadString(g_hInstance, IDS_ERROR_REGISTER_HOTKEY_FAILED, (LPTSTR)&lpFormatString, 0);

		// Simple message, doesn't have arguments
		lpMessageBuffer = _tcsdup(lpFormatString);
		break;

	default:
		// Unhandled error!
		assert(FALSE);
		lpMessageBuffer = _tcsdup(TEXT("Unexpected error condition"));
		break;
	}

	if (lpMessageBuffer != NULL)
	{
		// Show message box to user
		MessageBox(g_hMainWindow, lpMessageBuffer, szCaption, MB_ICONERROR);

		// Free memory
		free(lpMessageBuffer);
		lpMessageBuffer = NULL;
	}

	va_end(arg_ptr);
}
